unit VirtualTreesHelper;

// The contents of this file are subject to the Mozilla Public License
// Version 1.1 (the "License"); you may not use this file except in compliance
// with the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/
//
// Alternatively, you may redistribute this library, use and/or modify it under the terms of the
// GNU Lesser General Public License as published by the Free Software Foundation;
// either version 2.1 of the License, or (at your option) any later version.
// You may obtain a copy of the LGPL at http://www.gnu.org/copyleft/.
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
// specific language governing rights and limitations under the License.
//
//
//
// Extension for the TVirtualStringTree ( https://github.com/TurboPack/VirtualTreeView )
// Created by Roman G. Ish ( roman@elwinsoft.com ) http://roman-ish.ru
// This extension provides extended cell selection functionality
//
// v1.0 26-Jun-2019 First public release
//

interface

uses Classes, SysUtils,  Generics.Collections, VirtualTrees;

type
  TCell = record
    Row, Column: Integer;
  end;

  TCellList = class(TList<TCell>);

  TTreeSelection = class
  private
    FOtherButtonTrigger: boolean; // other mouse button clicked
   // min and max columns and row numbers selected by cells list
    FMinListSelectedRow: Integer;
    FMaxListSelectedRow: Integer;
    FMinListSelectedCol: Integer;
    FMaxListSelectedCol: Integer;
//
    FSelectedCells: TCellList;
    FUnSelecedCells: TCellList;
    FVirtualTree: TVirtualStringTree;
    FClickedRow: Integer;
    FClickedCol: Integer;
    FFirstSelectionBlockColumn: Integer;
    FLastSelectionBlockColumn: Integer;
    FFirstSelectionBlockRow: Integer;
    FLastSelectionBlockRow: Integer;
    FIsSelectionActive: boolean;
    FIsMouseSelecting: boolean;
    FIsKeySelecting: boolean;
    function GetFirstSelectionBlockRow: Integer; inline;
    function GetFirstSelectionBlockColumn: Integer; inline;
    function GetLastSelectionBlockColumn: Integer; inline;
    function GetLastSelectionBlockRow: Integer;
    function GetFirstColumn: Integer;
    function GetLastColumn: Integer;
    function GetRowsSelected: Integer;
    procedure InvalidateTree;
    procedure SetClickedCell(const ARow, AColumn: Integer);
    procedure ExpandSelection(const ARow, AColumn: Integer);
    function IsInSelectionBlock(const ARow, AColumn: Integer): boolean; inline;
    procedure AddToSelectedCellsList(const ACell: TCell);
    function GetFirstRow: Integer;
    function GetLastRow: Integer;
  public
    constructor Create(ATree: TVirtualStringTree);
    destructor Destroy; override;
    procedure Clear;
    ///<summary>Checking if cell is in selection</summary>
    function IsInSelection(ARow, AColumn: Integer): boolean;
    function IsRowInSelection(ARow: Integer): boolean; inline;
    function IsRowInBlockSelection(ARow: Integer): boolean; inline;
    function GetRowColFromMouse(const AX, AY: Integer; var Row, Col: Integer): boolean;
    procedure ExpandLeft;
    procedure ExpandRight;
    procedure ExpandUp;
    procedure ExpandDown;
    procedure StartSelection(ARow, AColumn: Integer);
    procedure MouseSelectionFromClickedTo(X, Y: Integer);
    procedure StartSelectionIfMoved(ARow, AColumn: Integer);

    ///<summary>set necessary tree options</summary>
    procedure AdjustTree;

    procedure KeyPress(Key: Word; AShiftState: TShiftState);
    procedure MouseDown(X, Y: Integer; AShiftState: TShiftState);
    procedure MouseMove(X, Y: Integer; AShiftState: TShiftState);
    procedure MouseUp(X, Y: Integer; AShiftState: TShiftState);

    property VirtualTree: TVirtualStringTree read FVirtualTree;
    property IsSelectionActive: boolean read FIsSelectionActive;
    property IsMouseSelecting: boolean read FIsMouseSelecting;
    property FirstColumn: Integer read GetFirstColumn;
    property LastColumn: Integer read GetLastColumn write FLastSelectionBlockColumn;
    property FirstRow: Integer read GetFirstRow;
    property LastRow: Integer read GetLastRow;
    property RowsSelected: Integer read GetRowsSelected;
  end;

function VSNodeByIndex(Root: PVirtualNode; Index: Integer): PVirtualNode;
procedure VSSelectNode(Vs: TVirtualStringTree; Index: Integer);
function VSFocusedIndex(Vs: TVirtualStringTree): Integer;

implementation

uses System.Math;

constructor TTreeSelection.Create(ATree: TVirtualStringTree);
begin
  inherited Create;

  FSelectedCells := TCellList.Create;
  FUnSelecedCells := TCellList.Create;

  FVirtualTree := ATree;
  AdjustTree;
end;

destructor TTreeSelection.Destroy;
begin
  FSelectedCells.Free;
  FUnSelecedCells.Free;

  inherited Destroy;
end;


function TTreeSelection.GetRowColFromMouse(const AX, AY: Integer; var Row, Col: Integer): boolean;
var
  HitInfo: THitInfo;
begin
  Result := false;

  if FVirtualTree <> nil then
  begin
    FVirtualTree.GetHitTestInfoAt(AX, AY, true, HitInfo);
    if HitInfo.HitNode <> nil then
    begin
      Row := HitInfo.HitNode.Index;
      Col := HitInfo.HitColumn;
      Result := true;
    end;
  end;
end;

procedure TTreeSelection.ExpandLeft;
begin
  if (FFirstSelectionBlockColumn <= FLastSelectionBlockColumn) then
  begin
    if (FFirstSelectionBlockColumn > 0) then
    begin
      Dec(FFirstSelectionBlockColumn);
      InvalidateTree;
    end;
  end
  else
    if (FLastSelectionBlockColumn >= FFirstSelectionBlockColumn) then
      if FLastSelectionBlockColumn > 0 then
      begin
        Dec(FLastSelectionBlockColumn);
        InvalidateTree;
      end;
end;

procedure TTreeSelection.ExpandRight;
begin
  if (FLastSelectionBlockColumn >= FFirstSelectionBlockColumn) then
  begin
    Inc(FLastSelectionBlockColumn);
    InvalidateTree;
  end
  else
    if (FLastSelectionBlockColumn <= FFirstSelectionBlockColumn) then
      if FFirstSelectionBlockColumn > 1 then
      begin
        Dec(FFirstSelectionBlockColumn);
        InvalidateTree;
      end;
end;

procedure TTreeSelection.ExpandUp;
begin
  if (FFirstSelectionBlockRow <= FLastSelectionBlockRow) then
  begin
    if FFirstSelectionBlockRow > 1 then
    begin
      Dec(FFirstSelectionBlockRow);
      InvalidateTree;
    end;
  end
  else
    if (FFirstSelectionBlockRow >= FLastSelectionBlockRow) then
      if (FFirstSelectionBlockRow > 1) then
      begin
        Dec(FFirstSelectionBlockRow);
        InvalidateTree;
      end;
end;

procedure TTreeSelection.ExpandDown;
begin
  if (FFirstSelectionBlockRow <= FLastSelectionBlockRow) then
  begin
    Inc(FLastSelectionBlockRow);
    InvalidateTree;
  end
  else
    if (FFirstSelectionBlockRow >= FLastSelectionBlockRow) then
    begin
      Inc(FFirstSelectionBlockRow);
      InvalidateTree;
    end;
end;

procedure TTreeSelection.InvalidateTree;
begin
  if FVirtualTree <> nil then
  begin
    FVirtualTree.Invalidate;
  end;
end;

procedure TTreeSelection.AdjustTree;
begin
  if FVirtualTree <> nil then
  begin
    FVirtualTree.TreeOptions.SelectionOptions := FVirtualTree.TreeOptions.SelectionOptions +
      [toDisableDrawSelection, toExtendedFocus, toMultiSelect];

    FVirtualTree.TreeOptions.SelectionOptions := FVirtualTree.TreeOptions.SelectionOptions -
      [toFullRowSelect];

    FVirtualTree.TreeOptions.PaintOptions := FVirtualTree.TreeOptions.PaintOptions + [toAlwaysHideSelection];
    FVirtualTree.TreeOptions.PaintOptions := FVirtualTree.TreeOptions.PaintOptions - [toHideFocusRect];
  end;
end;

procedure TTreeSelection.KeyPress(Key: Word; AShiftState: TShiftState);
begin
  if (FVirtualTree = nil) or (Key <= $20) then
    Exit;

  if IsSelectionActive then
    if (AShiftState = []) then
    begin
      Clear;
      SetClickedCell(FVirtualTree.FocusedNode.Index, FVirtualTree.FocusedColumn);
    end;

  if (AShiftState = []) and IsSelectionActive then
    Clear
  else
    if (AShiftState = [ssShift]) then
    begin
      if IsSelectionActive then
      begin
        case Key of
          37: ExpandLeft;
          38: ExpandUp;
          39: ExpandRight;
          40: ExpandDown;
        end;
      end
      else
      begin
        if FVirtualTree.FocusedNode <> nil then
          case Key of
            37:
            begin
              StartSelection(FVirtualTree.FocusedNode.Index, FVirtualTree.FocusedColumn);
              FIsKeySelecting := true;
              ExpandLeft;
            end;

            38:
            begin
              StartSelection(FVirtualTree.FocusedNode.Index, FVirtualTree.FocusedColumn);
              FIsKeySelecting := true;
              ExpandUp;
            end;

            39:
            begin
              StartSelection(FVirtualTree.FocusedNode.Index, FVirtualTree.FocusedColumn);
              FIsKeySelecting := true;
              ExpandRight;
            end;

           40:
            begin
              StartSelection(FVirtualTree.FocusedNode.Index, FVirtualTree.FocusedColumn);
              FIsKeySelecting := true;
              ExpandDown;
            end;
          end;
      end;
    end
    else
      if (AShiftState = [ssCtrl]) then
      begin
        if Key = Ord('A') then
         if FVirtualTree.RootNodeCount > 0 then
          begin
            FSelectedCells.Clear;
            FUnSelecedCells.Clear;

            StartSelection(0, 0);
            FLastSelectionBlockRow := FVirtualTree.RootNodeCount - 1;
            if FVirtualTree.Header.Columns.Count > 0 then
              FLastSelectionBlockColumn := FVirtualTree.Header.Columns.Count - 1
            else
              FLastSelectionBlockColumn := 0;

            InvalidateTree;
          end;
      end;
end;

procedure TTreeSelection.MouseDown(X, Y: Integer; AShiftState: TShiftState);
var
  Row, Col, r: Integer;
  Cell, ClickedCell: TCell;
begin
  if not (ssLeft in AShiftState) then
  begin
    FOtherButtonTrigger := true;
    exit;
  end;

  if GetRowColFromMouse(X, Y, Row, Col) then
  begin
    if ssShift in AShiftState then
    begin
      if not FIsSelectionActive  then
        StartSelection(FClickedRow, FClickedCol);

      ExpandSelection(Row, Col);
      InvalidateTree;
    end
    else
      if ssCtrl in AShiftState then
      begin
        FIsSelectionActive := true;
        Cell.Row := Row;
        Cell.Column := Col;

        if IsInSelectionBlock(Row, Col) then
        begin
          // Clicked on cell selected by range
          // add / remove from an exception list

          if FUnSelecedCells.Remove(Cell) < 0 then
            FUnSelecedCells.Add(Cell);

          InvalidateTree;
        end
        else
        begin
         // out of range cell is clicked
          if FSelectedCells.Count = 0 then
          begin
            ClickedCell.Row := FClickedRow;
            ClickedCell.Column := FClickedCol;
            AddToSelectedCellsList(ClickedCell);
          end;

          AddToSelectedCellsList(Cell);
          InvalidateTree;
        end;
      end
      else
      begin
        Clear;
        SetClickedCell(Row, Col);
      end;
  end;
end;

procedure TTreeSelection.MouseMove(X, Y: Integer; AShiftState: TShiftState);
var
  Row, Col: Integer;
begin
  if FIsMouseSelecting then
  begin
    if GetRowColFromMouse(X, Y, Row, Col) then
    begin
      ExpandSelection(Row, Col);
      InvalidateTree;
    end;
  end
  else
    if GetRowColFromMouse(X, Y, Row, Col) and (ssLeft in AShiftState) and
       not (ssCtrl in AShiftState) and not (ssShift in AShiftState) then
    begin
      StartSelectionIfMoved(Row, Col);
      InvalidateTree;
    end;
end;

procedure TTreeSelection.MouseUp(X, Y: Integer; AShiftState: TShiftState);
begin
  if FOtherButtonTrigger then
  begin
    FOtherButtonTrigger := false;
    exit;
  end;

  if FVirtualTree <> nil then
  begin
    if FIsMouseSelecting  then
    begin
      FIsMouseSelecting := false; // mouse button released in selection mode
      exit;
    end;

    if (not (ssShift in AShiftState)) and (not (ssCtrl in AShiftState)) and IsSelectionActive then
      Clear;
  end;
end;

procedure TTreeSelection.Clear;
begin
  FSelectedCells.Clear;
  FUnSelecedCells.Clear;

  FIsSelectionActive := false;
  FIsMouseSelecting := false;
  FIsKeySelecting := false;

  FMinListSelectedRow := -1;
  FMaxListSelectedRow := -1;
  FMinListSelectedCol := -1;
  FMaxListSelectedCol := -1;


  FFirstSelectionBlockColumn := -1;
  FLastSelectionBlockColumn := -1;
  FFirstSelectionBlockRow := -1;
  FLastSelectionBlockRow := -1;

  InvalidateTree;
end;

function TTreeSelection.GetFirstSelectionBlockRow: Integer;
begin
  Result := Min(FFirstSelectionBlockRow, FLastSelectionBlockRow);
end;

function TTreeSelection.GetLastSelectionBlockRow: Integer;
begin
  Result := Max(FFirstSelectionBlockRow, FLastSelectionBlockRow);
end;

function TTreeSelection.GetFirstSelectionBlockColumn: Integer;
begin
  Result := Min(FFirstSelectionBlockColumn, FLastSelectionBlockColumn);
end;

function TTreeSelection.GetLastSelectionBlockColumn: Integer;
begin
  Result := Max(FFirstSelectionBlockColumn, FLastSelectionBlockColumn);
end;


function TTreeSelection.GetFirstColumn: Integer;
var
  FC: Integer;
begin
  FC := GetFirstSelectionBlockColumn;

  if FC >= 0 then
  begin
    if FMinListSelectedCol >= 0 then
      Result := Min(FC, FMinListSelectedCol)
    else
      Result := FC;
  end
  else
    Result := FMinListSelectedCol;
end;

function TTreeSelection.GetLastColumn: Integer;
var
  FC: Integer;
begin
  FC := GetLastSelectionBlockColumn;

  if FC >= 0 then
  begin
    if FMaxListSelectedCol >= 0 then
      Result := Max(FC, FMaxListSelectedCol)
    else
      Result := FC;
  end
  else
    Result := FMaxListSelectedCol;
end;

function TTreeSelection.GetRowsSelected: Integer;
begin
  Result := (GetLastSelectionBlockRow - GetFirstSelectionBlockRow) or
            (FMaxListSelectedRow - FMinListSelectedRow);
end;

procedure TTreeSelection.StartSelectionIfMoved(ARow, AColumn: Integer);
begin
  if (FClickedRow <> ARow) or (FClickedCol <> AColumn) then
  begin
    StartSelection(FClickedRow, FClickedCol);
    FIsMouseSelecting := true;
  end;
end;

procedure TTreeSelection.SetClickedCell(const ARow, AColumn: Integer);
begin
  FClickedRow := ARow;
  FClickedCol := AColumn;
end;

function TTreeSelection.IsRowInBlockSelection(ARow: Integer): boolean;
begin
  Result := InRange(ARow, FFirstSelectionBlockRow, FLastSelectionBlockRow);
end;

function TTreeSelection.IsRowInSelection(ARow: Integer): boolean;
begin
  Result := InRange(ARow, FFirstSelectionBlockRow, FLastSelectionBlockRow) or
            InRange(ARow, FMinListSelectedRow, FMaxListSelectedRow);
end;

function TTreeSelection.IsInSelection(ARow, AColumn: Integer): boolean;
var
  Cl: TCell;
begin
  Result := IsInSelectionBlock(ARow, AColumn);

  Cl.Row := ARow;
  Cl.Column := AColumn;

  if Result then
  begin
    if FUnSelecedCells.IndexOf(Cl) >= 0 then
      Result := false;
  end
  else
  begin
    if FSelectedCells.IndexOf(Cl) >= 0 then
      Result := true;
  end;
end;

procedure TTreeSelection.StartSelection(ARow, AColumn: Integer);
begin
  if FVirtualTree <> nil then
  begin
    FFirstSelectionBlockRow := ARow;
    FFirstSelectionBlockColumn := AColumn;
    FLastSelectionBlockRow := ARow;
    FLastSelectionBlockColumn := AColumn;
    FIsSelectionActive := true;
  end;
end;

procedure TTreeSelection.MouseSelectionFromClickedTo(X, Y: Integer);
var
  HitInfo: THitInfo;
begin
  if FVirtualTree <> nil then
  begin
    FVirtualTree.GetHitTestInfoAt(X, Y, true, HitInfo);
    if HitInfo.HitNode <> nil then
    begin
      FFirstSelectionBlockRow := FClickedRow;
      FFirstSelectionBlockColumn := FClickedCol;
      FLastSelectionBlockRow := HitInfo.HitNode.Index;
      FLastSelectionBlockColumn := HitInfo.HitColumn;
      FIsMouseSelecting := false;
      FIsSelectionActive := true;

      InvalidateTree;
    end;
  end;
end;

procedure TTreeSelection.ExpandSelection(const ARow, AColumn: Integer);
begin
  FLastSelectionBlockRow := ARow;
  FLastSelectionBlockColumn := AColumn;
end;

function VSNodeByIndex(Root: PVirtualNode; Index: Integer): PVirtualNode;
begin
  if Root <> nil then
  begin
    Result := Root.FirstChild;
    while (Result <> nil) and (Result.Index <> Index) do
      Result := Result.NextSibling;
    Exit;
  end;

  Result := nil;
end;

function TTreeSelection.GetFirstRow: Integer;
begin
  if FMinListSelectedRow >= 0 then
  begin
    if FFirstSelectionBlockRow >= 0 then
      Result := Min(FMinListSelectedRow, FFirstSelectionBlockRow)
    else
      Result := FMinListSelectedRow;
  end
  else
    Result := FFirstSelectionBlockRow;
end;

function TTreeSelection.GetLastRow: Integer;
begin
  if FMaxListSelectedRow >= 0 then
  begin
    if FLastSelectionBlockRow >= 0 then
      Result := Max(FMaxListSelectedRow, FLastSelectionBlockRow)
    else
      Result := FMaxListSelectedRow;
  end
  else
    Result := FLastSelectionBlockRow;
end;

procedure TTreeSelection.AddToSelectedCellsList(const ACell: TCell);

  //       -
  procedure _AddCellToList;
  begin
    FSelectedCells.Add(ACell);
    FMinListSelectedRow := Min(FMinListSelectedRow, ACell.Row);
    FMaxListSelectedRow := Max(FMaxListSelectedRow, ACell.Row);
    FMinListSelectedCol := Min(FMinListSelectedCol, ACell.Column);
    FMaxListSelectedCol := Max(FMaxListSelectedCol, ACell.Column);
  end;

var
  Cell: TCell;
  i: Integer;
begin
  if FSelectedCells.Count > 0 then
  begin
    // trying to remove from the list of cells
    if FSelectedCells.Remove(ACell) >= 0 then
    begin
      // cell was removed
      // recalculate min-max of line and column
      if FSelectedCells.Count >= 0 then
      begin
        Cell := FSelectedCells[0];
      	FMinListSelectedRow := ACell.Row;
        FMaxListSelectedRow := FMinListSelectedRow;
        FMinListSelectedCol := ACell.Column;
        FMaxListSelectedCol := FMinListSelectedCol;

        for i := 1 to FSelectedCells.Count - 1 do
        begin
          FMinListSelectedRow := Min(FMinListSelectedRow, Cell.Row);
          FMaxListSelectedRow := Max(FMaxListSelectedRow, Cell.Row);
          FMinListSelectedCol := Min(FMinListSelectedCol, Cell.Column);
          FMaxListSelectedCol := Max(FMaxListSelectedCol, Cell.Column);
        end;
      end;
    end
    else
    begin
      // cell is not found in selected cells list
      // add cell to the list and update min-max
      _AddCellToList;
    end;
  end
  else
  begin
    FSelectedCells.Add(ACell);
    FMinListSelectedRow := ACell.Row;
    FMaxListSelectedRow := ACell.Row;
    FMinListSelectedCol := ACell.Column;
    FMaxListSelectedCol := ACell.Column;
  end;
end;

function TTreeSelection.IsInSelectionBlock(const ARow, AColumn: Integer): boolean;
begin
  Result := InRange(ARow, FFirstSelectionBlockRow, FLastSelectionBlockRow) and
            InRange(AColumn, FFirstSelectionBlockColumn, FLastSelectionBlockColumn);
end;

procedure VSSelectNode(Vs: TVirtualStringTree; Index: Integer);
var
  Node: PVirtualNode;
begin
  Node := VSNodeByIndex(Vs.RootNode, Index);
  Vs.ClearSelection;
  Vs.Selected[Node] := True;
  Vs.Refresh;
end;

function VSFocusedIndex(Vs: TVirtualStringTree): Integer;
begin
  if Vs.FocusedNode <> nil then
    Result := Vs.FocusedNode.Index
  else
    Result := -1;
end;

end.
